#!/usr/bin/perl
#################################################################################################
# 
# Author: Eric Kollmann
# Version : 0.1
# Date : 2010-06-12
#
# Copyright 2010 Eric Kollmann (xnih13@gmail.com)
#
#
#################################################################################################
# based off pcapcat  by Kristinn Gudjonsson but now heavily modified....
#
# original Header info below, same copyright and other info applies
#################################################################################################
#  pcapcat
#################################################################################################
# This script reads a PCAP file and prints out all the connections in the file and gives
# the user the option of dumping the content of the TCP stream
#
# Author: Kristinn Gudjonsson
# Version : 0.1b
# Date : 17/08/09
#
# Copyright 2009 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


use strict;
use Getopt::Long; # read parameters
use Pod::Usage;
use Net::Pcap;
use NetPacket::Ethernet;
use NetPacket::IP;
use NetPacket::TCP;
use NetPacket::UDP;


# version information
my $version = '0.1';

# other variables (do not edit unless you know what you are doing)
my ($out,$file);
my $filter;
my $filter_c;
my $print_help;
my $options;
my ($ether,$ip,$trans,$seqnum,$acknum,$ipid);
my $err;
my $packets;
my $index;
my $show_version;
my $dump_nr;
my $only_syn = 1;
my $all;
my $i = 0;
my $c = 0;
my $s_index = 1;
my @conversations = ([],[]);
my $count;
my $tempFlag;
my %header;  #packet header info

# read options
GetOptions(
        "filter:s"=>\$filter,
        "read:s"=>\$file,
        "options:s"=>\$options,
        "version!"=>\$show_version,
        "help|?!"=>\$print_help
) or pod2usage( 2 );

# check if we are asking for help
pod2usage(1) if $print_help;

# print versioning information
show_version() if $show_version;

# check if file exists
pod2usage(2) unless -e $file;

# reset the index value
$index = 1;

# start reading the file
$packets = Net::Pcap::open_offline( $file, \$err );
if( defined $err )
{
	die 'Unable to read the PCAP file ' . $err . "\n";
}


# now we need to check if we are going to read all packets or dump a certain traffic
{
	# normal run, print all the packets

	# check if there is a filter defined
	$s_index = 0 if defined $filter;	# if we have a filter defined, then we cannot show the index value FYI

	# create the filter
	Net::Pcap::compile( $packets, \$filter_c, $filter, 1, undef );
	Net::Pcap::setfilter( $packets, $filter_c );

	# read all the packets
	Net::Pcap::loop( $packets, \%header, \&read_all_packets, '' );

        if ($options eq 'conv')
        {
                &print_conv_data;
        }
        elsif ($options eq 'ipinfo')
        {
                &print_ipinfo_data;
        }

	# close the network file (since we have finished our processing)
	Net::Pcap::close( $packets );
}

sub print_ipinfo_data
{
	print "IP and TCP packet info\n";
	for ($count = 0; $count < $c; $count++)
	{
                #just print out the ones that were syn/rstack combo's for now
		$tempFlag = $conversations[$count][0];
		if ($tempFlag eq "S-RA")
		{
			print "$conversations[$count][10] $conversations[$count][1]:$conversations[$count][2] -> $conversations[$count][3]:$conversations[$count][4] $conversations[$count][5] $conversations[$count][6] \n";
		}
	}

}

sub print_conv_data
{
        # start the output stuff
	my $fullconv = 0;
	my $fullconvclosed = 0;
	my $attemptconv = 0;
	my $partialconv = 0;
	my $partialconvtorndown = 0;
	for ($count = 0; $count < $c; $count++)
	{
		$tempFlag = $conversations[$count][0];
		if ($tempFlag eq "S-SA-A")     #3 way handshake
		{
			$fullconv++;
		}
		if ($tempFlag eq "S-SA-FA")     #partial converstation torn down
		{
			$partialconvtorndown++;
		}
		if ($tempFlag eq "S-SA-A-FA")  #3 way handshake that was closed
		{
			$fullconvclosed++;
		}
		if ($tempFlag eq "S-RA")       #attempted conversation that was Reset
		{
			$attemptconv++;
		}
#		if ($tempFlag eq "RA")         #converstation that we just see a reset for, no initial syn
#		{
#			$partialconv++;
#		}
#		if ($tempFlag eq "A")         #converstation that we just see an ack for no initial syn
#		{
#			$partialconv++;
#		}

	}

	print "Attempted conversations:  $attemptconv\n";
#	for ($count = 0; $count < $c; $count++)
#	{
#		$tempFlag = $conversations[$count][0];
#		if ($tempFlag eq "S-RA")
#		{
#			print "\t[$conversations[$count][9]] - Starts at packet number $conversations[$count][8]\n";
#		}
#	}

#	print "Partial conversations we don't have all of:  $partialconv\n";
#	for ($count = 0; $count < $c; $count++)
#	{
#		$tempFlag = $conversations[$count][0];
#		if ($tempFlag eq "A")
#		{
#			print "\t[$conversations[$count][9]] - Starts at packet number $conversations[$count][8]\n";
#		}
#	}

	print "partial conversations, but torn down???:  $partialconvtorndown\n";
	for ($count = 0; $count < $c; $count++)
	{
		$tempFlag = $conversations[$count][0];
		if ($tempFlag eq "S-SA-FA")
		{
			print "\t[$conversations[$count][9]] - \n";
                        print "\t\tStarts at packet number $conversations[$count][8] \n";
                        print "\t\tStart Time: $conversations[$count][10] \n";
                        print "\t\tClient Info: $conversations[$count][1]:$conversations[$count][2] -> $conversations[$count][3]:$conversations[$count][4]\n";
		}
	}

	print "3 way handshake conversations, but no closure:  $fullconv\n";
	for ($count = 0; $count < $c; $count++)
	{
		$tempFlag = $conversations[$count][0];
		if ($tempFlag eq "S-SA-A")
		{
			print "\t[$conversations[$count][9]] - \n";
                        print "\t\tStarts at packet number $conversations[$count][8] \n";
                        print "\t\tStart Time: $conversations[$count][10] \n";
                        print "\t\tClient Info: $conversations[$count][1]:$conversations[$count][2] -> $conversations[$count][3]:$conversations[$count][4]\n";
		}
	}

	print "3 way handshake conversations, with closure:  $fullconvclosed\n";
	for ($count = 0; $count < $c; $count++)
	{
		$tempFlag = $conversations[$count][0];
		if ($tempFlag eq "S-SA-A-FA")
		{
			print "\t[$conversations[$count][9]] - \n";
                        print "\t\tStarts at packet number $conversations[$count][8] \n";
                        print "\t\tStart Time: $conversations[$count][10] \n";
                        print "\t\tEnd Time: $conversations[$count][11] \n";
                        print "\t\tClient Info: $conversations[$count][1]:$conversations[$count][2] -> $conversations[$count][3]:$conversations[$count][4]\n";
		}
	}

	$index--;  #was incremented on the last packet and started at 1.
	print "Total Packets:  $index \n";
}


sub read_all_packets
{
	my $user_data = shift;
	my $header = shift;
	my $pack = shift;
	my $fcheck;
	my $input;
	
	my ($tempFlag, $tempSrcIP, $tempSrcPort, $tempDestIP, $tempDestPort, $tempSeqNum, $tempAckNum);
	my $count;

	# strip header information
	$ether = NetPacket::Ethernet->decode( $pack );	
	
	# check if IP
	#$8100 = VLAN, then check VLAN Protocol for $9000 = IP  (also known as EAPS), this should be done, but isn't at this time
	#$0800 = IPv4 
	#$86DD = IPv6, don't care about IPv6 at this time
	if( $ether->{type} eq 2048 )
	{
		$ip = NetPacket::IP->decode( $ether->{'data'} );

		# check if TCP 
		if( $ip->{'proto'} eq 6 )
		{
			# TCP
			$trans = NetPacket::TCP->decode( $ip->{'data'} );
			$seqnum =$trans->{'seqnum'};
			$acknum = $trans->{'acknum'};
			
			# IP
			$ipid = $ip->{'id'};

				# we don't care about ECN bits
				$fcheck = $trans->{'flags'} & 0x3f;

				# check if we have a SYN packet
				if( $fcheck == 0x02  )
				{
					$conversations[$c] = [];
					$conversations[$c][0] = "S";
					$conversations[$c][1] = "$ip->{'src_ip'}";
					$conversations[$c][2] = "$trans->{'src_port'}";
					$conversations[$c][3] = "$ip->{'dest_ip'}";
					$conversations[$c][4] = "$trans->{'dest_port'}";
					$conversations[$c][5] = "$ipid";
					$conversations[$c][6] = "$seqnum";
					$conversations[$c][7] = "$acknum";
					$conversations[$c][8] = "$index";  #Start index
					$conversations[$c][9] = "$c";  #Converstation Number
					$conversations[$c][10] = localtime($header->{'tv_sec'});     #need a better way to just strip time off of, don't need date info

					# increment both the index variable as well as the (for printing)
					$index++;
					$c++;
				}
				# check if we have a ACK
				elsif( $fcheck == 0x10  )
				{
                                        #loop through @conversations and see if we are part of one of them
                                        my $found = 0;
					for ($count = 0; $count < $c; $count++)
					{
						$tempFlag = $conversations[$count][0];

						if ($tempFlag eq 'S-SA')
						{
							$tempSrcIP = $conversations[$count]->[1];
							$tempSrcPort = $conversations[$count]->[2];
							$tempDestIP = $conversations[$count]->[3];
							$tempDestPort = $conversations[$count]->[4];
							$tempSeqNum = $conversations[$count]->[6];
							$tempAckNum = $conversations[$count]->[7];

							if (($ip->{'src_ip'} eq $tempSrcIP) &
							   ($trans->{'src_port'} eq $tempSrcPort) &
						           ($ip->{'dest_ip'} eq $tempDestIP) &
						           ($trans->{'dest_port'} eq $tempDestPort) &
						           ($seqnum eq $tempSeqNum + 1))
						        {
								$conversations[$count][0] = "$tempFlag-A";
								$found = 1;
							}
						}
					}
#					if ($found eq 0) #didn't find it, so partial conversation
#					{
#                                                my $ackFound = 0;
#                                                for ($count = 0; $count < $c; $count++)
#					        {
#						        $tempFlag = $conversations[$count][0];
#
#						        if ($tempFlag eq 'A')
#						        {
#                                                                $tempSrcIP = $conversations[$count]->[1];
#							        $tempSrcPort = $conversations[$count]->[2];
#							        $tempDestIP = $conversations[$count]->[3];
#							        $tempDestPort = $conversations[$count]->[4];
#							        $tempSeqNum = $conversations[$count]->[6];
#							        $tempAckNum = $conversations[$count]->[7];
#
#							        if (($ip->{'src_ip'} eq $tempSrcIP) &
#							           ($trans->{'src_port'} eq $tempSrcPort) &
#						                   ($ip->{'dest_ip'} eq $tempDestIP) &
#						                   ($trans->{'dest_port'} eq $tempDestPort) &
#						                   ($seqnum eq $tempSeqNum + 1))
#                                                               {
#								        $conversations[$count][0] = "$tempFlag-A";
#								        $ackFound = 1;
#                                                               }
#                                                       }
#                                                }
#                                               if ($ackFound eq 0)
#                                               {
#		                                	$conversations[$c] = [];
#                     					$conversations[$c][0] = "A";
#                          				$conversations[$c][1] = "$ip->{'src_ip'}";
#					                $conversations[$c][2] = "$trans->{'src_port'}";
#                     					$conversations[$c][3] = "$ip->{'dest_ip'}";
#                                    			$conversations[$c][4] = "$trans->{'dest_port'}";
#                                      			$conversations[$c][5] = "$ipid";
#                                            		$conversations[$c][6] = "$seqnum";
#					                $conversations[$c][7] = "$acknum";
#                     					$conversations[$c][8] = "$index";  #Start index
#	                            			$conversations[$c][9] = "$c";  #Converstation Number
#                          				$conversations[$c][10] = localtime($header->{'tv_sec'});
#                             				$conversations[$c][11] = "$StopTime";
#
#			                		# increment both the index variable as well as the (for printing)
#					               $index++;
#                                         	       $c++;
#                                               }
#                                        }
					$index++;
				}
				# check if we have a FIN/ACK
				# technically I should check for a FA in the other direction also, but....
				elsif( $fcheck == 0x11  )
				{
                                        #loop through @conversations and see if we are part of one of them
					for ($count = 0; $count < $c; $count++)
					{
						$tempFlag = $conversations[$count][0];

						if ($tempFlag eq 'S-SA-A')
						{
							$tempSrcIP = $conversations[$count]->[1];
							$tempSrcPort = $conversations[$count]->[2];
							$tempDestIP = $conversations[$count]->[3];
							$tempDestPort = $conversations[$count]->[4];
							$tempSeqNum = $conversations[$count]->[6];
							$tempAckNum = $conversations[$count]->[7];
						
							if (($ip->{'src_ip'} eq $tempSrcIP) &
							   ($trans->{'src_port'} eq $tempSrcPort) & 
						           ($ip->{'dest_ip'} eq $tempDestIP) &
						           ($trans->{'dest_port'} eq $tempDestPort))
						        {
								$conversations[$count][0] = "$tempFlag-FA";
                            					$conversations[$count][11] = localtime($header->{'tv_sec'});  #for now ignoring tv_nsec
							}
						}

						if ($tempFlag eq 'S-SA')
						{
							$tempSrcIP = $conversations[$count]->[1];
							$tempSrcPort = $conversations[$count]->[2];
							$tempDestIP = $conversations[$count]->[3];
							$tempDestPort = $conversations[$count]->[4];
							$tempSeqNum = $conversations[$count]->[6];
							$tempAckNum = $conversations[$count]->[7];
						
							if (($ip->{'src_ip'} eq $tempSrcIP) &
							   ($trans->{'src_port'} eq $tempSrcPort) &
						           ($ip->{'dest_ip'} eq $tempDestIP) &
						           ($trans->{'dest_port'} eq $tempDestPort) &
						           ($seqnum eq $tempSeqNum + 1))
						        {
								$conversations[$count][0] = "$tempFlag-FA";
                            					$conversations[$count][11] = localtime($header->{'tv_sec'});  #for now ignoring tv_nsec
							}
						}

					}
					$index++;
				}
				# check if we have a SYN/ACK
				elsif( $fcheck == 0x12  )
				{
                                        #loop through @conversations and see if we are part of one of them
					for ($count = 0; $count < $c; $count++)
					{
						$tempFlag = $conversations[$count][0];
						
						if ($tempFlag eq 'S')
						{
							$tempSrcIP = $conversations[$count]->[1];
							$tempSrcPort = $conversations[$count]->[2];
							$tempDestIP = $conversations[$count]->[3];
							$tempDestPort = $conversations[$count]->[4];
							$tempSeqNum = $conversations[$count]->[6];
							$tempAckNum = $conversations[$count]->[7];
						
							if (($ip->{'src_ip'} eq $tempDestIP) &
							   ($trans->{'src_port'} eq $tempDestPort) & 
						           ($ip->{'dest_ip'} eq $tempSrcIP) &
						           ($trans->{'dest_port'} eq $tempSrcPort) &
						           ($acknum eq $tempSeqNum + 1))
						        {
								$conversations[$count][0] = "$tempFlag-SA";
							}
						}
					}
					$index++;
				}
				# check if we have a RST/ACK
				elsif( $fcheck == 0x14  )
				{
                                        #loop through @conversations and see if we are part of one of them
					for ($count = 0; $count < $c; $count++)
					{
						$tempFlag = $conversations[$count][0];
						
						if ($tempFlag eq 'S')
						{
							$tempSrcIP = $conversations[$count]->[1];
							$tempSrcPort = $conversations[$count]->[2];
							$tempDestIP = $conversations[$count]->[3];
							$tempDestPort = $conversations[$count]->[4];
							$tempSeqNum = $conversations[$count]->[6];
							$tempAckNum = $conversations[$count]->[7];
						
							if (($ip->{'src_ip'} eq $tempDestIP) &
							   ($trans->{'src_port'} eq $tempDestPort) & 
						           ($ip->{'dest_ip'} eq $tempSrcIP) &
						           ($trans->{'dest_port'} eq $tempSrcPort) &
						           ($acknum eq $tempSeqNum + 1))
						        {
								$conversations[$count][0] = "$tempFlag-RA";
							}  
							#we may have RA packets from other conversations that we did not have the Syn packet for, I am not tracking them at this time.
							#evidence 6 has 120 RA packets, 1 of them is packet 1343 which is a Reset packet between it and port 8080 which we didn't initially see
						}
					}
					$index++;
				}
				# check if we have a PSH/ACK, may not need
				elsif( $fcheck == 0x18  )
				{
					$index++;
				}
				else
				{
					$index++;
				}
				
	
		}
		else #udp, icmp, non tcp, so just bypass
		{
			$index++;
			$i++;
		}
	}
	else  #non IPv4 packets, arp, cdp, stuff like that, just bypass them
	{
          $index++;
          $i++;
        }

		
}

# a simple sub routine to show 
sub show_version
{
	print $0,' version ',$version, ' copyright 2010, Eric Kollmann',"\n";
	exit 0;
}

0;

__END__

=pod

=head1 NAME

B<pcapcat> - a simple script to read PCAP file and dump the content of the conversation into a file

=head1 SYNOPSIS 

B<pcapcat> [-f|--filter PCAP FILTER] [-o|--options conv ipinfo] -r|--read PCAP_FILE

B<pcapcat> [-v|--version] [-h|--help|-?]

=head1 OPTIONS

=over 8

=item B<-f|-filter PCAP_FILTER>

Enter a traditional PCAP filter to filter out the content of the file (see man tcpdump for further details about constructing such a filter)

=item B<-r|-read PCAP_FILE>

The PCAP file that the script should read

=item B<-v|-version>

Dump the version number of the script to the screen and quit

=item B<-h|-help|-?>

Print this help menu

=back
